require 'set'
require 'find'
require 'digest/md5'
require 'yaml'
require 'sadelta'
require 'stringio'
require 'fileutils'
require 'fastcst/ui'
require 'zlib'

include SuffixArrayDelta


module ChangeSet

    # = Introduction
    # 
    # ChangeSet performs an analysis of a source and target directory and then allows you
    # to write the changeset to two files.  To apply a ChangeSet you simply use the 
    # ChangeSet#apply_changeset function.  Applying a changeset is incredibly simple because
    # of the way a changeset file is designed.
    #
    # The initialize function does most of the work, but leaves the moved files detection to
    # detect_moved_files.  This is done since moved files detection is really optional, and may
    # not be requested.
    #
    # = Design
    #
    # A ChangeSet object performs an analysis of the source and target directory.  It then
    # writes a YAML dump of a series of Operation objects (MoveOperation, DeltaOperation,
    # DeleteOperation, and CreateOperation) and write necessary data to a raw data output.
    # This makes creating a changeset file incredibly easy and makes it easy to create new
    # operations.  Once all the operations are written to disk then the changeset is finished.
    #
    # Operations which need to store data use a given data output stream (which writes to a file)
    # and keep track of how much they write.  When an Operation is applied it then knows
    # how much to read from the data stream to do its job.  The consequence of this is that
    # the operations must be run in order.  Each operation supports a skip method in order
    # to allow it to be skipped.
    #
    # Applying a changeset becomes incredibly easy then:  just load each operation from the
    # YAML file and call its run method.  The run method knows what it needs to do, and only
    # needs the directory to do it in and the data stream to read stuff from (if necessary).
    #
    # One key thing is that, if skip is called, and the operation expects to read a certain 
    # amount of data from the data stream, then it should seek ahead that much so the next
    # operation is not messed up.
    #
    # = Usage
    #
    # For more information on how it is used refer to MakeChangeSetCommand#run.
    #
    # = Future Directions
    #
    #   * It doesn't record directory change information yet.  This is coming soon.
    #   * No crypto yet.  That's a requirement in the future.
    #
    # = Weirdness
    #
    # One thing that you will notice is that this class uses the UI module in ui.rb to give
    # all it's presentation to the user.  This module is an abstraction away from the actual
    # display method which will eventually be designed to support different UI systems 
    # depending on user preference or build options.
    #

    class ChangeSetBuilder
        attr_reader :deleted, :created, :common, :moved, :changed
    
        # Does the majority of the change detection using the Set class.  It basically
        # scans both source and target, and then determines the deleted, created, and common
        # files in each by full pathname (this does not include directories yet).  Then
        # it scans the common files to see which ones have changed mtimes and assumes these
        # are changed.  Even if the files didn't really change, they did have an attribute
        # (mtime) change so they need to be recorded.  The DeltaOperation class will
        # figure out what really changed and record appropriately.
        # 
        # Moved file detection is done with detect_moved_files since this is optional.
        def initialize(source, target)
            @source = source
            @target = target

            src_files, src_dirs = scan(@source)
            tgt_files, tgt_dirs = scan(@target)

            # use Set to figure out what has possibly changed
            @deleted = src_files - tgt_files
            @created = tgt_files - src_files
            @common = src_files & tgt_files
            @moved = {}  # initially empty until dected_moved_files is requested
            @changed = {}
            
            # directories are handled after everything else
            @deleted_dirs = src_dirs - tgt_dirs
            @created_dirs = tgt_dirs - src_dirs
            
            @common.each do |file|
                otime = File.mtime(File.join(source,file))
                ntime = File.mtime(File.join(target,file))
                if otime != ntime
                    # file has changed, but we if the old file is 0 length then its actually a create
                    if File.size(File.join(source,file)) == 0
                        @created << file
                    else
                        @changed[file] = [otime, ntime]
                    end
                end
            end
        end
    
    
        # Performs a simple moved file detection operation.  A moved file is defined as:
        # 
        #   1.  Has same file name.
        #   2.  Has different base path.
        #   3.  Has same file size.
        #   4.  Has same md5 digest.
        #
        # The algorithm first indexes the basenames of the @deleted and @created files and then
        # goes through the deleted files and created files looking for basenames with only 1 element
        # in the value array.  When it finds one then it checks their sizes and md5 digests.  If
        # everything checks out then it records this as a move and removes the files from the
        # @created and @deleted lists.
    
        def detect_moved_files
            md5sums = {}  # use as a reverse lookup
        
            del_basenames = index_base_names(@deleted)
            tgt_basenames = index_base_names(@created)
            
            del_basenames.each do |basename, files|
                # we only process files that have unique locations mentioned, this is a direct possible move
                tgt_files = tgt_basenames[basename]
                if tgt_files && tgt_files.length == 1 && files.length == 1
                    del_digest = ""
                    tgt_digest = ""
                    from_file = files[0]
                    to_file = tgt_files[0]
                    
                    # next test is to simply compare files sizes, can't be same file if different size
                    if File.size?(@source + '/' + from_file) == File.size?(@target + '/' + to_file)
                        
                        # now generate the hashes for both files as the final confirmation of same file
                        Dir.chdir(@source) { del_digest = Digest::MD5.digest(File.read(from_file)) }
                        Dir.chdir(@target) { tgt_digest = Digest::MD5.digest(File.read(to_file)) }
                        
                        if del_digest == tgt_digest
                            # digests match so the basenames are the same and the digests are the same, it's a move
                            @moved[from_file] = [to_file, File.stat(File.join(@target,to_file)).mtime]
                            
                            # and clean the files from either side to eliminate them
                            @deleted.delete from_file
                            @created.delete to_file
                        end
                    end
                end
            end
        end
    
    
        # Returns true if there are detected changes.
        def has_changes?
            @deleted.size > 0 || @created.size > 0 || @changed.size > 0 || @moved.size > 0
        end

        
    
        # Write journal and data to the two output streams.  It basically
        # just creates the correct operation objects in order and writes
        # then to the journal output stream as a series of YAML documents.
        def write_changeset(journal_out, data_out)
        
            @deleted.sort.each do |path|
                YAML.dump(DeleteOperation.new(path), journal_out)
                journal_out.write("\n")
            end
        
            @moved.sort.each do |from, to_info|
                YAML.dump(MoveOperation.new(from, to_info[0], to_info[1]), journal_out)
                journal_out.write("\n")
            end
        
            @created.sort.each do |path|
                YAML.dump(CreateOperation.new(@target, path, data_out), journal_out)
                journal_out.write("\n")
            end
        
            @changed.sort.each do |file, info|
                YAML.dump(DeltaOperation.new(@source, @target, file, data_out), journal_out)
                journal_out.write("\n")
            end
            
            # finally we write a DirectoryOperation that is responsible for intelligently
            # deleting any directories which no longer exist in the target
            YAML.dump(DirectoryOperation.new(@deleted_dirs, @created_dirs), journal_out)
            journal_out.write("\n")
        end
    
    
    
        private
        # Private method that does an inverse indexing of the basenames in the given hash.
        def index_base_names(filenames)
            basenames = {}
            filenames.each do |list|
                list.each do |file|
                    base = File.basename file
                    basenames[base] ||=[]
                    basenames[base] << file
                end
            end
        
            return basenames
        end
    
        # Scans a directory for all the files, but skips anything that isn't a file or directory
        # It also skips unix style "hidden" files which start with a "." so that it avoids
        # files that usually aren't wanted.  There is a real need to create an excludes
        # list of some sort for this.
        # 
        # It returns two Set objects, one for the files and one for the directories it finds.
        def scan(dir)
            file_results = Set.new
            dir_results = Set.new
            
            Dir.chdir(dir) do
                Find.find(".") do |file|
                    base = File.basename(file)

                    # TODO: allow for configurable exclude lists
                    if base != "." and base[0,1] == "."
                        if File.directory?(file)
                            # skip this "hidden" directory
                            Find.prune
                            # otherwise we just ignore the hidden file
                        end
                    elsif File.file? file
                        file_results << file
                    elsif File.directory? file
                        dir_results << file
                    end
                end
            end
            
            return file_results, dir_results
        end
    end



    # Analyzes the journal file (input IO) and produces a hash with some statistics
    # in it.
        def ChangeSet.statistics(journal_in)
            stats = {"moves" => 0, "creates" => 0, "deletes" => 0, "deltas" => 0}
            # no need to run skip since we're not doing anything other than counting them
            YAML.each_document(journal_in) do |op|
                case op
                when DeleteOperation:
                    stats["deletes"] += 1
                when CreateOperation:
                    stats["creates"] += 1
                when MoveOperation:
                    stats["moves"] += 1
                when DeltaOperation:
                    stats["deltas"] += 1
                when DirectoryOperation:
                    stats["deleted directories"] = op.deleted_dirs
                    stats["created directories"] = op.created_dirs
                else
                    UI.failure :input, "Unknown operation #{op.class}"
                end
            end
        
            return stats
        end

        
        # A function that reads a journal and data input stream,
        # and runs the operations against the target directory.
        # It closes the input streams for the caller to ensure that
        # everything is cleaned up properly.
        #
        # Setting test_run==true will run all operations in test mode where they don't actually
        # do anything to the target directory, but will report on any errors and return
        # true/false if they work or not.
        def ChangeSet.apply_changeset(journal_in, data_in, target_dir, test_run=false)
            begin
                failure_count = 0
                
                YAML.each_document(journal_in) do |op|
                    # add one to the failure count unless the operation runs fine
                    if test_run
                        failure_count += 1 unless op.test(data_in, target_dir)
                    else
                        failure_count += 1 unless op.run(data_in, target_dir)
                    end
                end
            ensure
                journal_in.close if journal_in
                data_in.close if data_in
            end
            
            return failure_count
        end
    

        # A utility method to easily create a changeset given just the
        # changeset name (it adds the ".yaml" and ".fcs" for the journal
        # and data files).  It returns the ChangeSetBuilder for you to
        # analyze, and it will not make the changeset if there are
        # no changes reported.
        def ChangeSet.make_changeset(cs_name, source, target)
            changes = ChangeSetBuilder.new(source, target)
            
            if not changes.has_changes?
                UI.event :exit, "Nothing changed.  Exiting."
            else
                begin
                    changes.detect_moved_files
                
                    md_out = Zlib::GzipWriter.new(File.open(cs_name + ".yaml", "w"))
                    data_out = Zlib::GzipWriter.new(File.open(cs_name + ".fcs", "w"))
                
                    changes.write_changeset(md_out, data_out)
                ensure
                    md_out.close
                    data_out.close
                end
            end
            
            return changes
        end
        
    
    
        # A small utility function that safely creates a path leading up to a given file.
        # It does this by splitting off the base path with File#split, and then creating the
        # path.
        #
        # If the path exists, then it checks to see if it is a directory.  In the event that the
        # target path is not a directory it will delete it and then re-create it as such.  This
        # means that if you have a file there, then the file will be destroyed.
        def ChangeSet.create_target_path(filename)
            path = File.split(filename)[0]
    
            if File.exist? path
                # target path exists, but sometimes the target path is a file
                # this doesn't get picked up by the delete operations so we delete it here
                # and recreate the file as a directory for further operations
                if not File.directory? path
                    UI.event :warn, "#{path} exists but is not a directory, deleting and recreating as a directory (will backup with ~ ending)"
                    # create a backup of the file with a ~ ending
                    FileUtils::mv(path, path +"~")
                    FileUtils::rm_rf(path)
                
                    UI.event :create, "creating path: #{path}"
                    FileUtils::mkdir_p(path)
                end
            else
                UI.event :create, "creating path: #{path}"
                FileUtils::mkdir_p(path)
            end
        end
    


    # An operation which deletes a given file.  It only needs the path
    # during initialization, and just blindly deletes the file when run.
    class DeleteOperation    
    
        # Records the path that needs to be deleted.
        def initialize(path)
            UI.event :delete, path
            @path = path
        end
    
    
        # Changes to the dir and deletes the file.
        def run(indata, dir)
            begin
                Dir.chdir dir do
                    UI.event :delete, @path
                    FileUtils.rm @path
                end
            rescue
                UI.failure :delta, "#{$!}"
                return false
            end
            
            return true
        end
        
        # Performs tests to see if it's safe/possible to perform this operation
        def test(indata, dir)
            begin
                Dir.chdir dir do
                    FileUtils::DryRun.rm @path
                end
                
            rescue
                UI.failure :delta, "#{$!}"
                skip(indata)
                return false
            end
            
            return true
        end
        
    
        # Skip for delete does nothing.
        def skip(indata)
            UI.event :skip, "DELETE #@path"
        end
    end



    # An operation for creating files based on the target directory, path to the file,
    # and a data output stream.  It records the file's contents and the mtime of the
    # file so that it can replicate the file as closely as possible.
    class CreateOperation
    
        # Records the file's path, mtime, and contents.  The dir argument
        # is used only to get to the file, and isn't recorded in the object.
        def initialize(dir, path, out)
            @path = path
        
            Dir.chdir dir do
                @symlink = File.symlink? @path
                @mtime = File.mtime path
            
                if @symlink
                    # it's a symlink so record the target path
                    UI.event :symlink, "Storing #@path as a symlink rather than a new file"
                    @symlink_target = File.readlink(@path)
                else
                    # regular file so go to town
                    data = File.read @path
                    @length = data.length
                    UI.event :create, "#@path #@length"
                    out.write data
                end
            end
        end
    
    
        # Creates the file in the target directory, reading the data out of the
        # indata stream.  After the file is created this function will set the
        # mtime so it matches the original.  This will overwrite the file  it
        # it already exists, so it isn't safe yet.
        def run(indata, dir)
            UI.event :create, @path

            begin
                Dir.chdir dir do
                    ChangeSet.create_target_path(@path)
            
                    if @symlink
                        UI.event :symlink, "Creating symlink to #@symlink_target"
                        # as usual, symlink is backwards
                        File.symlink(@symlink_target, @path)
                    else
                        UI.event :create, @path
                        data = indata.read(@length)
                        
                        File.open(@path, "wb") {|out| out.write data }
                    end
                    
                    UI.event :mtime, "Fixing mtime for #@path"
                    File.utime(Time.now, @mtime, @path)
                end
            rescue
                UI.failure :create, "Failed to create file: #$!"
                return false
            end
            
            return true
        end
    
        # Performs tests to see if it's safe/possible to perform this operation
        def test(indata, dir)
            begin
                Dir.chdir dir do
                    # the path exists, so we can test this one
                    if File.exist?(File.dirname(@path))
                        FileUtils::DryRun.touch(@path)
                    end
                end
            rescue
                UI.failure :create, "Failed to create file: #$!"
                skip(indata) # skip any data we need
                return false
            end
            
            return true
        end

        
        # Seeks ahead the @length.
        def skip(indata)
            UI.event :skip, "CREATE #@path"
            indata.seek(@length, IO::SEEK_CUR)
        end
    end


    # Records a move operation that moves a file from one path to another,
    # and then sets it's mtime to the original setting.
    class MoveOperation
    
        # It needs to know the from and to paths, and the target mtime.
        def initialize(from_path, to_path, mtime)
            @from_path = from_path
            @to_path = to_path
            @mtime = mtime
            UI.event :move, "#@from_path -> #@to_path"
        end

        # Simply performs the move and the mtime set.
        def run(indata, dir)
            UI.event :move, "#@from_path -> #@to_path"
            begin
                Dir.chdir dir do
                    ChangeSet.create_target_path(@to_path, test)
                    FileUtils.mv @from_path, @to_path, :force => true
                    File.utime(Time.now, @mtime, @to_path)
                end
                
                skip(indata) # skip any data we need
            rescue
                UI.failure :move, "#{$!}"
                return false
            end
            
            return true
        end
    
        # Performs tests to see if it's safe/possible to perform this operation
        def test(indata, dir)
            begin
                Dir.chdir dir do
                    return false unless File.exist?(@from_path)
                    
                    if File.exist?(File.dirname(@to_path))
                        FileUtils::DryRun.mv @from_path, @to_path, :force => true
                    end
                end
            rescue
               UI.failure :move, "#{$!}"
               skip(indata)
               return false
            end
            
            return true
        end
    
        # Skip does nothing.
        def skip(indata)
            UI.event :skip, "MOVE #@from_path -> #@to_path"
        end
    end



    # The most complicated operation, it performs a delta between two files
    # and records the delta, it's length, and the mtime.  It will avoid recording
    # a delta if the file's contents haven't changed.
    class DeltaOperation
    
        def initialize(source, target, path, out)
            @path = path
        
            # we just ignore symlinks since the real change happens in the target file
            @symlink =  File.symlink? @path
        
            # don't do anything if its a symlink
            if @symlink
                UI.event :warn, "Deltas against symlinks are ignored since they are pointless"
            else @symlink        
                # setup the remaining journal
                @mtime = File.mtime(File.join(target, path))
            
                # read the gear and do the delta
                src_data = File.read(File.join(source, path))
                tgt_data = File.read(File.join(target, path))
                
                io_out = StringIO.new
            
                match_count, match_total, insert_count, insert_total = SuffixArrayDelta::make_delta(src_data, tgt_data, io_out)
            
                # get the md5 hash of the delta and the target data before we write it out
                if (src_data.length == tgt_data.length and match_count == 1 and match_total == src_data.length and insert_count == 0 and insert_total == 0)
                    # no changes actually, so we configure this command that with a 0 length
                    @length = 0
                    UI.event :same, @path
                else
                    # there's actual changes so write it out after recording the length
                    UI.event :delta, @path
                
                    @length = io_out.pos
                    io_out.rewind
                    out.write io_out.read
                end
            end
        end
    
    
        # Run reads in the delta, changes to the dir, and then applies the delta
        # in order to create the changed file.  It will skip a file if it's missing.
        def run(indata, dir)
        
            # don't bother running if this is a symlink
            if @symlink
                UI.event :symlink, "Skipped delta of symlink #@path"
            else
                begin
                    Dir.chdir dir do 
                        UI.event :delta, @path
                        
                        # skip the file if it doesn't exist
                        if not File.exist? @path
                            UI.event :missing,  "Can't delta #@path, file missing"
                            return false
                        end
                        
                        # no need to process the the delta if it's 0 length
                        if @length > 0
                            reference = File.read @path
                            delta = StringIO.new(indata.read(@length))
                            outfile = File.open(@path, "wb")

                            SuffixArrayDelta::apply_delta(reference, delta, outfile)
                            
                            outfile.close
                        end
                
                        # update the file's mtime
                        File.utime(Time.new, @mtime, @path)
                    end
                rescue
                    UI.failure :delta, "#$!"
                    outfile.close if outfile
                    return false
                end
            end
            
            return true
        end
    
        # Performs tests to see if it's safe/possible to perform this operation
        def test(indata, dir)
            begin
                Dir.chdir do
                    return true if @symlink
                    return false unless File.exist? @path
                    
                    # do a test run of the delta on the file contents
                    if @length > 0
                        delta = StringIO.new(indata.read(@length))
                        reference = File.read @path
                        outfile = StringIO.new
                        
                        SuffixArrayDelta::apply_delta(reference, delta, outfile)
                    end
                end
            rescue
                UI.failure :delta, "#$!"
                # no need to skip here since we actually read during this test
                return false
            end
            
            return true
        end
        
    
        # Skips ahead in the data stream by the @length
        def skip(indata)
            UI.event :skip, "DELTA #@path"
            indata.seek(@length, IO::SEEK_CUR)
        end
    end

    

    # Responsible for deleting the directories which don't exist anymore
    # but doing it so that there is no damage if the person has some files 
    # in the directory.
    class DirectoryOperation    
    
        attr_reader :deleted_dirs
        attr_reader :created_dirs
        
        # Records the path that needs to be deleted.
        def initialize(deleted_dirs, created_dirs)
            @created_dirs = created_dirs.sort
            @deleted_dirs = deleted_dirs.sort.reverse
        end
    
    
        # Changes to the dir and deletes the file.
        def run(indata, dir)
            begin
                Dir.chdir dir do
                    @deleted_dirs.each do |path|
                        begin
                            UI.event :rmdir, path
                            FileUtils.rmdir path
                        rescue
                            UI.failure :directory, "Could not remove directory: #$!"
                        end
                    end
                    
                    @created_dirs.each do |path|
                        if not File.exist? path
                            UI.event :mkdir, path
                            FileUtils.mkdir_p path
                        end
                    end
                end
            rescue
                UI.failure :directory, "#{$!}"
                return false
            end
            
            return true
        end
        
        # Performs tests to see if it's safe/possible to perform this operation
        def test(indata, dir)
            begin
                Dir.chdir dir do
                    @deleted_dirs.each do |path|
                        FileUtils::DryRun.rmdir path
                    end
                    
                    @created_dirs.each do |path|
                        FileUtils::DryRun.mkdir_p path
                    end
                end
                
            rescue
                UI.failure :delta, "#{$!}"
                skip(indata)
                return false
            end
            
            return true
        end
        
    
        # Skip for delete does nothing.
        def skip(indata)
            UI.event :skip, "Directory deletes"
        end
    end
    
end    
