require 'fastcst/ui'
require 'fastcst/distrib'


module Repository

    class FinishCommand < Command
        def initialize(argv)
            super(argv, [
                ["-s", "--show", "Show the current log", :@show]
            ])
            
            @repo_dir = Repository.search
        end
        
        def validate
            valid? @repo_dir, "Could not find the repository directory"
            
            return @valid
        end
        
        def run
            repo = Repository.new @repo_dir
            
            # check that a revision is in progress
            if not repo['Current Revision']
                UI.failure :constraint, "You have not used begin to start a revision yet."
                return
            end
            
            parent_id = repo['Path'].pop
            md_file = File.join(repo.work_dir, "meta-data.yaml")
            md = MetaData.load_metadata(md_file)
            cs_name = repo['Project'] + '-' + md['Revision']
            
            Dir.chdir repo.work_dir do
                originals = File.join("..","originals")
                sources = File.join("..","..")
                
                data_file = cs_name + ".fcs"
                journal_file = cs_name + ".yaml"
                
                changes = ChangeSet.make_changeset(cs_name, originals, sources)

                # abort if there were no changes
                if not changes.has_changes?
                    return
                end
                
                # create the undo
                if not File.exist? "undo.yaml"
                    # the undo changeset has to go in the reverse direction
                    UI.start_finish("Creating 'undo' revision") do
                        changes = ChangeSet.make_changeset("undo", sources, originals)
                    end
                end
                
                # sync the originals directory
                journal_in = Zlib::GzipReader.new(File.open(journal_file))
                data_in = Zlib::GzipReader.new(File.open(data_file))
                
                # apply closes the files for us
                ChangeSet.apply_changeset(journal_in, data_in, originals)

                # create the meta-data
                MetaData.finish_metadata(md_file, parent_id, data_file, journal_file)
            end
        
            # store the newly created changeset, doing a move instead of a copy
            md = repo.store_changeset repo.work_dir, "meta-data.yaml", move=true
            
            # get the new location and move the undo there
            cs_path, md = repo.find_changeset(md['ID'])
            FileUtils.mv(File.join(repo.work_dir, "undo.yaml"), cs_path)
            FileUtils.mv(File.join(repo.work_dir, "undo.fcs"), cs_path)
            
            # and update the environment to reflect our new revision path
            repo["Path"] = repo["Path"] << md['ID']
            
            UI.event :finished, "New revision: #{repo.build_readable_name md['ID']}"
            
            # and remove the current revision
            repo.delete 'Current Revision'
        end
    
    end
end
